from .base import (
    BaseEntityRepository,
    BaseReadOnlyEntityRepository,
    BaseRepositoryFactory,
    BaseWriteOnlyEntityRepository,
)

__all__ = (
    "BaseEntityRepository",
    "BaseReadOnlyEntityRepository",
    "BaseRepositoryFactory",
    "BaseWriteOnlyEntityRepository",
)



import typing as t
from dataclasses import dataclass

from .exceptions import AttributeCannotBeOverriddenError

PARAMS = t.TypeVar("PARAMS", bound="BaseParams")


@dataclass(frozen=True)
class BaseParams:
    pass


class BaseContext(t.Protocol[PARAMS]):
    params: PARAMS


@dataclass
class BaseMutableContext(BaseContext[PARAMS]):
    params: PARAMS

    def __setattr__(self, name: str, value: t.Any) -> None:
        if hasattr(self, name):
            raise AttributeCannotBeOverriddenError(name)

        super().__setattr__(name, value)


@dataclass(frozen=True)
class BaseImmutableContext(BaseContext[PARAMS]):
    params: PARAMS



import typing as t
from dataclasses import dataclass, field

from .....datapoint import (
    BaseDataPointContainer,
    BaseDataPointSpec,
    DataPointCannotBeOverriddenError,
    DataPointIsNotProducedError,
)
from ..base import BaseDataPointContext, BaseParams

PARAMS = t.TypeVar("PARAMS", bound=BaseParams)


@dataclass(frozen=True)
class InMemoryDataPointContext(
    BaseDataPointContext[
        PARAMS, BaseDataPointContainer[t.Any], BaseDataPointSpec[t.Any]
    ]
):
    params: PARAMS

    _storage: t.Dict[BaseDataPointSpec[t.Any], t.Any] = field(
        init=False, repr=False, default_factory=dict
    )

    async def add(self, *datapoints: BaseDataPointContainer[t.Any]) -> None:
        for datapoint in datapoints:
            if datapoint.spec in self._storage:
                raise DataPointCannotBeOverriddenError(datapoint.spec)
            else:
                self._storage[datapoint.spec] = datapoint.value

    async def get(
        self, *specs: BaseDataPointSpec[t.Any]
    ) -> t.Dict[BaseDataPointSpec[t.Any], t.Any]:
        missing = {name for name in specs if name not in self._storage}

        if len(missing) > 0:
            raise DataPointIsNotProducedError(missing)

        return {k: v for k, v in self._storage.items() if k in specs}



from .base import (
    BaseDataPointConsumer,
    BaseDataPointContainer,
    BaseDataPointProducer,
    BaseDataPointSpec,
)
from .exceptions import (
    DataPointCannotBeOverriddenError,
    DataPointIsNotProducedError,
)
from .impl import ConsumesDataPoints, ProducesDataPoints

__all__ = (
    "BaseDataPointContainer",
    "BaseDataPointConsumer",
    "BaseDataPointSpec",
    "BaseDataPointProducer",
    "ConsumesDataPoints",
    "ProducesDataPoints",
    "DataPointCannotBeOverriddenError",
    "DataPointIsNotProducedError",
)



import typing as t

from ....context import BaseContext
from ....result import Result
from ....unit.aio import BaseUnit
from ...aio import BaseUnitProxy, BaseWorkManager

CONTEXT = t.TypeVar("CONTEXT", bound=BaseContext[t.Any])
OUT = t.TypeVar("OUT")


class NoOpUnitProxy(BaseUnitProxy[CONTEXT, OUT]):  # pragma: no cover
    def __init__(
        self,
        *,
        unit: BaseUnit[CONTEXT, OUT],
    ) -> None:
        self._unit = unit

    async def __call__(self, context: CONTEXT) -> Result[OUT]:
        return await self._unit(context)


class NoOpWorkManager(BaseWorkManager):  # pragma: no cover
    def by(self, unit: BaseUnit[CONTEXT, OUT]) -> NoOpUnitProxy[CONTEXT, OUT]:
        return NoOpUnitProxy(unit=unit)



from .base import BaseTransaction, BaseTransactionManager
from .impl import TransactionalUnitProxy, TransactionalWorkManager

__all__ = (
    "BaseTransaction",
    "BaseTransactionManager",
    "TransactionalUnitProxy",
    "TransactionalWorkManager",
)



from .base import BaseDataPointConsumer, BaseDataPointProducer
from .impl import ConsumesDataPoints, ProducesDataPoints

__all__ = (
    "BaseDataPointConsumer",
    "BaseDataPointProducer",
    "ConsumesDataPoints",
    "ProducesDataPoints",
)



import typing as t
from abc import ABC, abstractmethod
from dataclasses import asdict

from ....clock import offset_naive_utcnow

try:
    from sqlalchemy import delete, exists, insert, select, update
except ImportError:  # pragma: no cover
    raise ImportError(
        "Seems that you are trying to import extra module that was not installed,"
        " please install pyuow[sqlalchemy]"
    )

from ....contrib.sqlalchemy.tables import (
    AuditedEntityTable,
    EntityTable,
    VersionedEntityTable,
)
from ....contrib.sqlalchemy.work.impl import (
    SqlAlchemyReadOnlyTransactionManager,
    SqlAlchemyTransactionManager,
)
from ....entity import AuditedEntity, Entity, VersionedEntity
from ....repository import BaseEntityRepository, BaseRepositoryFactory

ENTITY_ID = t.TypeVar("ENTITY_ID", bound=t.Any)
ENTITY_TYPE = t.TypeVar("ENTITY_TYPE", bound=Entity[t.Any])
ENTITY_TABLE = t.TypeVar("ENTITY_TABLE", bound=EntityTable)


class BaseSqlAlchemyEntityRepository(
    t.Generic[ENTITY_ID, ENTITY_TYPE, ENTITY_TABLE],
    BaseEntityRepository[ENTITY_ID, ENTITY_TYPE],
    ABC,
):
    def __init__(
        self,
        table: t.Type[ENTITY_TABLE],
        transaction_manager: SqlAlchemyTransactionManager,
        readonly_transaction_manager: SqlAlchemyReadOnlyTransactionManager,
    ) -> None:
        self._table = table
        self._transaction_manager = transaction_manager
        self._readonly_transaction_manager = readonly_transaction_manager

    @staticmethod
    @abstractmethod
    def to_entity(record: ENTITY_TABLE) -> ENTITY_TYPE:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_record(entity: ENTITY_TYPE) -> ENTITY_TABLE:
        raise NotImplementedError

    def find(self, entity_id: ENTITY_ID) -> t.Optional[ENTITY_TYPE]:
        statement = select(self._table).where(self._table.id == entity_id)

        with self._readonly_transaction_manager.transaction() as trx:
            result = (trx.it().execute(statement)).scalar_one_or_none()

        return self.to_entity(result) if result else None

    def find_all(
        self, entity_ids: t.Iterable[ENTITY_ID]
    ) -> t.Iterable[ENTITY_TYPE]:
        statement = select(self._table).where(self._table.id.in_(entity_ids))

        with self._readonly_transaction_manager.transaction() as trx:
            result = (trx.it().execute(statement)).scalars().all()

        return [self.to_entity(record) for record in result]

    def get(self, entity_id: ENTITY_ID) -> ENTITY_TYPE:
        statement = select(self._table).where(self._table.id == entity_id)

        with self._readonly_transaction_manager.transaction() as trx:
            result = (trx.it().execute(statement)).scalar_one()

        return self.to_entity(result)

    def exists(self, entity_id: ENTITY_ID) -> bool:
        statement = (
            exists(self._table).where(self._table.id == entity_id).select()
        )

        with self._readonly_transaction_manager.transaction() as trx:
            return (trx.it().execute(statement)).scalar() or False

    def add(self, entity: ENTITY_TYPE) -> ENTITY_TYPE:
        statement = (
            insert(self._table)
            .values(**asdict(self.to_record(entity)))
            .returning(self._table)
        )

        with self._transaction_manager.transaction() as trx:
            result = (trx.it().execute(statement)).scalar_one()

        return self.to_entity(result)

    def add_all(
        self, entities: t.Sequence[ENTITY_TYPE]
    ) -> t.Iterable[ENTITY_TYPE]:
        return [self.add(entity) for entity in entities]

    def update(self, entity: ENTITY_TYPE) -> ENTITY_TYPE:
        record = self.to_record(entity)
        conditions = []

        if (
            isinstance(record, AuditedEntityTable)
            and issubclass(self._table, AuditedEntityTable)
            and isinstance(entity, AuditedEntity)
        ):
            record.updated_date = offset_naive_utcnow()
            conditions.append(self._table.created_date == entity.created_date)
            conditions.append(self._table.updated_date == entity.updated_date)

        if (
            isinstance(record, VersionedEntityTable)
            and issubclass(self._table, VersionedEntityTable)
            and isinstance(entity, VersionedEntity)
        ):
            record.version = entity.version.next()
            conditions.append(self._table.version == entity.version)

        statement = (
            update(self._table)
            .values(**asdict(record))
            .where(self._table.id == entity.id, *conditions)
            .returning(self._table)
        )

        with self._transaction_manager.transaction() as trx:
            result = (trx.it().execute(statement)).scalar_one()

        return self.to_entity(result)

    def update_all(
        self, entities: t.Sequence[ENTITY_TYPE]
    ) -> t.Iterable[ENTITY_TYPE]:
        return [self.update(entity) for entity in entities]

    def delete(self, entity: ENTITY_TYPE) -> bool:
        statement = (
            delete(self._table)
            .where(self._table.id == entity.id)
            .returning(self._table.id)
        )

        with self._transaction_manager.transaction() as trx:
            identifier = (trx.it().execute(statement)).scalar_one()

        return t.cast(bool, identifier == entity.id)


class BaseSqlAlchemyRepositoryFactory(BaseRepositoryFactory, ABC):
    def __init__(
        self,
        transaction_manager: SqlAlchemyTransactionManager,
        readonly_transaction_manager: SqlAlchemyReadOnlyTransactionManager,
    ) -> None:
        self._transaction_manager = transaction_manager
        self._readonly_transaction_manager = readonly_transaction_manager



from .base import BaseUnit
from .exceptions import CannotReassignUnitError, FinalUnitError
from .impl import ConditionalUnit, ErrorUnit, FinalUnit, FlowUnit, RunUnit

__all__ = (
    "BaseUnit",
    "CannotReassignUnitError",
    "FinalUnitError",
    "ConditionalUnit",
    "ErrorUnit",
    "FinalUnit",
    "FlowUnit",
    "RunUnit",
)



import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass

VALUE = t.TypeVar("VALUE", bound=t.Any)


@dataclass(frozen=True)
class BaseDataPointSpec(t.Generic[VALUE]):
    name: str
    ref_type: t.Type[VALUE]

    def __call__(self, value: VALUE) -> "BaseDataPointContainer[VALUE]":
        return BaseDataPointContainer(self, value)


@dataclass(frozen=True)
class BaseDataPointContainer(t.Generic[VALUE]):
    spec: BaseDataPointSpec[VALUE]
    value: VALUE


DATA_POINT_CONTAINER = t.TypeVar(
    "DATA_POINT_CONTAINER", bound=BaseDataPointContainer[t.Any]
)


class BaseDataPointProducer(t.Generic[DATA_POINT_CONTAINER], ABC):
    @abstractmethod
    def add(self, *datapoints: DATA_POINT_CONTAINER) -> None:
        raise NotImplementedError


DATA_POINT_SPEC = t.TypeVar("DATA_POINT_SPEC", bound=BaseDataPointSpec[t.Any])


class BaseDataPointConsumer(t.Generic[DATA_POINT_SPEC], ABC):
    @abstractmethod
    def get(self, *specs: DATA_POINT_SPEC) -> t.Dict[DATA_POINT_SPEC, t.Any]:
        raise NotImplementedError



from .base import (
    AuditedEntityTable,
    BaseTable,
    EntityTable,
    VersionedEntityTable,
)

__all__ = (
    "AuditedEntityTable",
    "BaseTable",
    "EntityTable",
    "VersionedEntityTable",
)



import abc
import typing as t
from abc import ABC

from ..context import BaseContext
from ..result import Result

CONTEXT = t.TypeVar("CONTEXT", bound=BaseContext[t.Any])
OUT = t.TypeVar("OUT")


class BaseUnit(t.Generic[CONTEXT, OUT], ABC):
    @abc.abstractmethod
    def __call__(self, context: CONTEXT) -> Result[OUT]:
        raise NotImplementedError



import typing as t
from dataclasses import dataclass
from datetime import datetime

from ..clock import offset_naive_utcnow
from ..types import MISSING

ENTITY_ID = t.TypeVar("ENTITY_ID", bound=t.Any)
ENTITY_TYPE = t.TypeVar("ENTITY_TYPE", bound="Entity[t.Any]")


class Version(int):
    def __new__(cls, value: int) -> "Version":
        if value < 0:
            raise ValueError("Version cannot be negative")

        return super().__new__(cls, value)

    def next(self) -> "Version":
        return Version(self + 1)


@dataclass(frozen=True)
class Entity(t.Generic[ENTITY_ID]):
    id: ENTITY_ID


@dataclass(frozen=True)
class AuditedEntity(Entity[ENTITY_ID]):
    created_date: datetime = t.cast(datetime, MISSING)
    updated_date: datetime = t.cast(datetime, MISSING)

    def __post_init__(self) -> None:
        now = offset_naive_utcnow()

        if self.created_date is MISSING:
            object.__setattr__(self, "created_date", now)
        if self.updated_date is MISSING:
            object.__setattr__(self, "updated_date", now)


@dataclass(frozen=True)
class VersionedEntity(Entity[ENTITY_ID]):
    version: Version = Version(0)



from .base import (
    BaseContext,
    BaseImmutableContext,
    BaseMutableContext,
    BaseParams,
)
from .exceptions import AttributeCannotBeOverriddenError

__all__ = (
    "BaseContext",
    "BaseImmutableContext",
    "BaseMutableContext",
    "BaseParams",
    "AttributeCannotBeOverriddenError",
)



import abc
import typing as t
from abc import ABC

TRANSACTION_PROVIDER = t.TypeVar("TRANSACTION_PROVIDER")


class BaseTransaction(t.Generic[TRANSACTION_PROVIDER], ABC):
    def __init__(self, transaction_provider: TRANSACTION_PROVIDER) -> None:
        self._transaction_provider = transaction_provider

    def it(self) -> TRANSACTION_PROVIDER:
        return self._transaction_provider

    @abc.abstractmethod
    async def rollback(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def commit(self) -> None:
        raise NotImplementedError


TRANSACTION = t.TypeVar("TRANSACTION", bound=BaseTransaction[t.Any])


class BaseTransactionManager(t.Generic[TRANSACTION], ABC):
    @abc.abstractmethod
    def transaction(self) -> t.AsyncContextManager[TRANSACTION]:
        raise NotImplementedError



import typing as t
from abc import ABC, abstractmethod

from ..base import BaseDataPointContainer, BaseDataPointSpec

DATA_POINT_CONTAINER = t.TypeVar(
    "DATA_POINT_CONTAINER", bound=BaseDataPointContainer[t.Any]
)


class BaseDataPointProducer(t.Generic[DATA_POINT_CONTAINER], ABC):
    @abstractmethod
    async def add(self, *datapoints: DATA_POINT_CONTAINER) -> None:
        raise NotImplementedError


DATA_POINT_SPEC = t.TypeVar("DATA_POINT_SPEC", bound=BaseDataPointSpec[t.Any])


class BaseDataPointConsumer(t.Generic[DATA_POINT_SPEC], ABC):
    @abstractmethod
    async def get(
        self, *specs: DATA_POINT_SPEC
    ) -> t.Dict[DATA_POINT_SPEC, t.Any]:
        raise NotImplementedError



import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass

from ..datapoint import (
    BaseDataPointConsumer,
    BaseDataPointContainer,
    BaseDataPointProducer,
    BaseDataPointSpec,
    DataPointIsNotProducedError,
)


class ConsumesDataPoints(ABC):
    @property
    @abstractmethod
    def _consumes(
        self, *specs: BaseDataPointSpec[t.Any]
    ) -> t.Set[BaseDataPointSpec[t.Any]]:
        raise NotImplementedError

    def out_of(
        self, consumer: BaseDataPointConsumer[t.Any]
    ) -> t.Dict[BaseDataPointSpec[t.Any], t.Any]:
        return consumer.get(*self._consumes)


class ProducesDataPoints(ABC):
    @dataclass(frozen=True)
    class ProducerProxy:
        _producer: BaseDataPointProducer[t.Any]
        _required_names: t.Set[BaseDataPointSpec[t.Any]]

        def add(self, *datapoints: BaseDataPointContainer[t.Any]) -> None:
            actual_names = {datapoint.spec for datapoint in datapoints}
            missing = self._required_names - actual_names

            if len(missing) > 0:
                raise DataPointIsNotProducedError(missing)

            self._producer.add(*datapoints)

    @property
    @abstractmethod
    def _produces(
        self, *specs: BaseDataPointSpec[t.Any]
    ) -> t.Set[BaseDataPointSpec[t.Any]]:
        raise NotImplementedError

    def to(self, producer: BaseDataPointProducer[t.Any]) -> ProducerProxy:
        return self.ProducerProxy(producer, self._produces)



import abc
import typing as t
from abc import ABC
from logging import getLogger

from ..context import BaseContext
from ..result import Result
from ..types import MISSING, MissingType
from .base import BaseUnit
from .exceptions import CannotReassignUnitError, FinalUnitError

logger = getLogger(__name__)

CONTEXT = t.TypeVar("CONTEXT", bound=BaseContext[t.Any])
OUT = t.TypeVar("OUT")


class FlowUnit(BaseUnit[CONTEXT, OUT], ABC):
    def __init__(self) -> None:
        self._root: "FlowUnit[CONTEXT, OUT]" = self
        self._next: t.Union["FlowUnit[CONTEXT, OUT]", MissingType] = MISSING

    def __rshift__(
        self: "FlowUnit[CONTEXT, OUT]", other: "FlowUnit[CONTEXT, OUT]"
    ) -> "FlowUnit[CONTEXT, OUT]":
        if not isinstance(other._next, MissingType):
            raise CannotReassignUnitError

        self._next = other
        other._root = self._root
        return other

    def build(self) -> "FlowUnit[CONTEXT, OUT]":
        return self._root


class FinalUnit(FlowUnit[CONTEXT, OUT], ABC):
    def __call__(self, context: CONTEXT) -> Result[OUT]:
        cls_name = self.__class__.__name__

        try:
            result = self.finish(context)
        except Exception as error:
            logger.exception(
                "[%s] failed with exception", cls_name, exc_info=error
            )
            return Result.error(error)
        else:
            logger.info("[%s] completed", self.__class__.__name__)

        logger.debug("[%s] result [%s]", self.__class__.__name__, result)
        return result

    @abc.abstractmethod
    def finish(self, context: CONTEXT) -> Result[OUT]:
        raise NotImplementedError

    def __rshift__(
        self: "FlowUnit[CONTEXT, OUT]", other: "FlowUnit[CONTEXT, OUT]"
    ) -> "FlowUnit[CONTEXT, OUT]":
        raise FinalUnitError(self.__class__.__name__)


@t.final
class ErrorUnit(FinalUnit[CONTEXT, OUT]):
    def __init__(
        self,
        exc: Exception,
    ) -> None:
        super().__init__()
        self._exc = exc

    def finish(self, context: CONTEXT) -> Result[OUT]:
        return Result.error(self._exc)


class ConditionalUnit(FlowUnit[CONTEXT, OUT]):
    def __init__(self, *, on_failure: FlowUnit[CONTEXT, OUT]) -> None:
        super().__init__()
        self._on_failure = on_failure

    def __call__(self, context: CONTEXT) -> Result[OUT]:
        cls_name = self.__class__.__name__

        if isinstance(self._next, MissingType):
            raise NotImplementedError(f"[{cls_name}] next unit is not set")

        try:
            passed = self.condition(context)
        except Exception as error:
            logger.exception(
                "[%s] failed with exception", cls_name, exc_info=error
            )
            return Result.error(error)

        if passed:
            logger.info("[%s] completed", cls_name)
            return self._next(context)

        logger.info("[%s] failed", cls_name)
        return self._on_failure(context)

    @abc.abstractmethod
    def condition(self, context: CONTEXT) -> bool:
        raise NotImplementedError


class RunUnit(FlowUnit[CONTEXT, OUT]):
    def __call__(self, context: CONTEXT) -> Result[OUT]:
        cls_name = self.__class__.__name__

        if isinstance(self._next, MissingType):
            raise NotImplementedError(f"[{cls_name}] next unit is not set")

        try:
            self.run(context)
        except Exception as error:
            logger.exception(
                "[%s] failed with exception", cls_name, exc_info=error
            )
            return Result.error(error)

        logger.info("[%s] completed", cls_name)
        return self._next(context)

    @abc.abstractmethod
    def run(self, context: CONTEXT) -> None:
        raise NotImplementedError



import typing as t
from abc import ABC
from dataclasses import dataclass

from ...context import BaseContext
from ...datapoint import (
    BaseDataPointConsumer,
    BaseDataPointContainer,
    BaseDataPointProducer,
    BaseDataPointSpec,
)
from ..base import BaseParams

PARAMS = t.TypeVar("PARAMS", bound=BaseParams)


DATA_POINT_CONTAINER = t.TypeVar(
    "DATA_POINT_CONTAINER", bound=BaseDataPointContainer[t.Any]
)


@dataclass(frozen=True)
class BaseDataPointProducerContext(
    t.Generic[PARAMS, DATA_POINT_CONTAINER],
    BaseContext[PARAMS],
    BaseDataPointProducer[DATA_POINT_CONTAINER],
    ABC,
):
    pass


DATA_POINT_SPEC = t.TypeVar("DATA_POINT_SPEC", bound=BaseDataPointSpec[t.Any])


@dataclass(frozen=True)
class BaseDataPointConsumerContext(
    t.Generic[PARAMS, DATA_POINT_SPEC],
    BaseContext[PARAMS],
    BaseDataPointConsumer[DATA_POINT_SPEC],
    ABC,
):
    pass


@dataclass(frozen=True)
class BaseDataPointContext(
    t.Generic[PARAMS, DATA_POINT_CONTAINER, DATA_POINT_SPEC],
    BaseDataPointProducerContext[PARAMS, DATA_POINT_CONTAINER],
    BaseDataPointConsumerContext[PARAMS, DATA_POINT_SPEC],
    ABC,
):
    pass



import abc
import typing as t
from abc import ABC
from logging import getLogger

from ...context import BaseContext
from ...result import Result
from ...types import MISSING, MissingType
from ..exceptions import CannotReassignUnitError, FinalUnitError
from .base import BaseUnit

logger = getLogger(__name__)

CONTEXT = t.TypeVar("CONTEXT", bound=BaseContext[t.Any])
OUT = t.TypeVar("OUT")


class FlowUnit(BaseUnit[CONTEXT, OUT], ABC):
    def __init__(self) -> None:
        self._root: "FlowUnit[CONTEXT, OUT]" = self
        self._next: t.Union["FlowUnit[CONTEXT, OUT]", MissingType] = MISSING

    def __rshift__(
        self: "FlowUnit[CONTEXT, OUT]",
        other: "FlowUnit[CONTEXT, OUT]",
    ) -> "FlowUnit[CONTEXT, OUT]":
        if not isinstance(other._next, MissingType):
            raise CannotReassignUnitError

        self._next = other
        other._root = self._root
        return other

    def build(self) -> "FlowUnit[CONTEXT, OUT]":
        return self._root


class FinalUnit(FlowUnit[CONTEXT, OUT], ABC):
    async def __call__(self, context: CONTEXT) -> Result[OUT]:
        cls_name = self.__class__.__name__

        try:
            result = await self.finish(context)
        except Exception as error:
            logger.exception(
                "[%s] failed with exception", cls_name, exc_info=error
            )
            return Result.error(error)
        else:
            logger.info("[%s] completed", self.__class__.__name__)

        logger.debug("[%s] result [%s]", self.__class__.__name__, result)
        return result

    @abc.abstractmethod
    async def finish(self, context: CONTEXT) -> Result[OUT]:
        raise NotImplementedError

    def __rshift__(
        self: "FlowUnit[CONTEXT, OUT]",
        other: "FlowUnit[CONTEXT, OUT]",
    ) -> "FlowUnit[CONTEXT, OUT]":
        raise FinalUnitError(self.__class__.__name__)


@t.final
class ErrorUnit(FinalUnit[CONTEXT, OUT]):
    def __init__(
        self,
        exc: Exception,
    ) -> None:
        super().__init__()
        self._exc = exc

    async def finish(self, context: CONTEXT) -> Result[OUT]:
        return Result.error(self._exc)


class ConditionalUnit(FlowUnit[CONTEXT, OUT]):
    def __init__(self, *, on_failure: FlowUnit[CONTEXT, OUT]) -> None:
        super().__init__()
        self._on_failure = on_failure

    async def __call__(self, context: CONTEXT) -> Result[OUT]:
        cls_name = self.__class__.__name__

        if isinstance(self._next, MissingType):
            raise NotImplementedError(f"[{cls_name}] next unit is not set")

        try:
            passed = await self.condition(context)
        except Exception as error:
            logger.exception(
                "[%s] failed with exception", cls_name, exc_info=error
            )
            return Result.error(error)

        if passed:
            logger.info("[%s] completed", cls_name)
            return await self._next(context)

        logger.info("[%s] failed", cls_name)
        return await self._on_failure(context)

    @abc.abstractmethod
    async def condition(self, context: CONTEXT) -> bool:
        raise NotImplementedError


class RunUnit(FlowUnit[CONTEXT, OUT]):
    async def __call__(self, context: CONTEXT) -> Result[OUT]:
        cls_name = self.__class__.__name__

        if isinstance(self._next, MissingType):
            raise NotImplementedError(f"[{cls_name}] next unit is not set")

        try:
            await self.run(context)
        except Exception as error:
            logger.exception(
                "[%s] failed with exception", cls_name, exc_info=error
            )
            return Result.error(error)

        logger.info("[%s] completed", cls_name)
        return await self._next(context)

    @abc.abstractmethod
    async def run(self, context: CONTEXT) -> None:
        raise NotImplementedError



import abc
import typing as t
from abc import ABC

from ...entity import Entity

ENTITY_ID = t.TypeVar("ENTITY_ID", bound=t.Any)
ENTITY_TYPE = t.TypeVar("ENTITY_TYPE", bound=Entity[t.Any])


class BaseReadOnlyEntityRepository(t.Generic[ENTITY_ID, ENTITY_TYPE], ABC):
    @abc.abstractmethod
    async def find(self, entity_id: ENTITY_ID) -> t.Optional[ENTITY_TYPE]:
        raise NotImplementedError

    @abc.abstractmethod
    async def find_all(
        self, entity_ids: t.Iterable[ENTITY_ID]
    ) -> t.Iterable[ENTITY_TYPE]:
        raise NotImplementedError

    @abc.abstractmethod
    async def get(self, entity_id: ENTITY_ID) -> ENTITY_TYPE:
        raise NotImplementedError

    @abc.abstractmethod
    async def exists(self, entity_id: ENTITY_ID) -> bool:
        raise NotImplementedError


class BaseWriteOnlyEntityRepository(t.Generic[ENTITY_ID, ENTITY_TYPE], ABC):
    @abc.abstractmethod
    async def add(self, entity: ENTITY_TYPE) -> ENTITY_TYPE:
        raise NotImplementedError

    @abc.abstractmethod
    async def add_all(
        self, entities: t.Sequence[ENTITY_TYPE]
    ) -> t.Iterable[ENTITY_TYPE]:
        raise NotImplementedError

    @abc.abstractmethod
    async def update(self, entity: ENTITY_TYPE) -> ENTITY_TYPE:
        raise NotImplementedError

    @abc.abstractmethod
    async def update_all(
        self, entities: t.Sequence[ENTITY_TYPE]
    ) -> t.Iterable[ENTITY_TYPE]:
        raise NotImplementedError

    @abc.abstractmethod
    async def delete(self, entity: ENTITY_TYPE) -> bool:
        raise NotImplementedError


class BaseEntityRepository(
    BaseReadOnlyEntityRepository[ENTITY_ID, ENTITY_TYPE],
    BaseWriteOnlyEntityRepository[ENTITY_ID, ENTITY_TYPE],
    ABC,
):
    pass


class BaseRepositoryFactory(ABC):
    @property
    @abc.abstractmethod
    def repositories(
        self,
    ) -> t.Mapping[
        t.Type[Entity[t.Any]],
        BaseEntityRepository[t.Any, t.Any],
    ]:
        raise NotImplementedError

    def repo_for(
        self, entity_type: t.Type[ENTITY_TYPE]
    ) -> BaseEntityRepository[t.Any, t.Any]:
        try:
            return self.repositories[entity_type]
        except KeyError as e:
            raise KeyError(
                f"Repository for {entity_type.__name__} is not registered"
            ) from e



import typing as t
from asyncio import current_task
from contextlib import asynccontextmanager

try:
    from sqlalchemy.ext.asyncio import (
        AsyncEngine,
        AsyncSession,
        async_scoped_session,
        async_sessionmaker,
    )
except ImportError:  # pragma: no cover
    raise ImportError(
        "Seems that you are trying to import extra module that was not installed,"
        " please install pyuow[sqlalchemy]"
    )

from .....work.aio.transactional import BaseTransaction, BaseTransactionManager


class SqlAlchemyTransaction(BaseTransaction[AsyncSession]):
    async def rollback(self) -> None:
        await self._transaction_provider.rollback()

    async def commit(self) -> None:
        await self._transaction_provider.commit()


class SqlAlchemyTransactionManager(
    BaseTransactionManager[SqlAlchemyTransaction]
):
    def __init__(self, engine: AsyncEngine) -> None:
        self._session_factory = async_scoped_session(
            async_sessionmaker(engine, expire_on_commit=False),
            scopefunc=current_task,
        )

    @asynccontextmanager
    async def transaction(self) -> t.AsyncIterator[SqlAlchemyTransaction]:
        session = self._session_factory()
        if session.in_transaction():
            yield SqlAlchemyTransaction(session)
        else:
            async with session.begin():
                yield SqlAlchemyTransaction(session)


class SqlAlchemyReadOnlyTransactionManager(SqlAlchemyTransactionManager):
    def __init__(self, engine: AsyncEngine) -> None:
        super().__init__(
            engine.execution_options(isolation_level="AUTOCOMMIT")
        )

    @asynccontextmanager
    async def transaction(self) -> t.AsyncIterator[SqlAlchemyTransaction]:
        async with self._session_factory() as session:
            yield SqlAlchemyTransaction(session)



import typing as t
from contextlib import contextmanager

try:
    from sqlalchemy import Engine
    from sqlalchemy.orm import Session, scoped_session, sessionmaker
except ImportError:  # pragma: no cover
    raise ImportError(
        "Seems that you are trying to import extra module that was not installed,"
        " please install pyuow[sqlalchemy]"
    )

from ....work.transactional import BaseTransaction, BaseTransactionManager


class SqlAlchemyTransaction(BaseTransaction[Session]):
    def rollback(self) -> None:
        self._transaction_provider.rollback()

    def commit(self) -> None:
        self._transaction_provider.commit()


class SqlAlchemyTransactionManager(
    BaseTransactionManager[SqlAlchemyTransaction]
):
    def __init__(self, engine: Engine) -> None:
        self._session_factory = scoped_session(
            sessionmaker(engine, expire_on_commit=False)
        )

    @contextmanager
    def transaction(self) -> t.Iterator[SqlAlchemyTransaction]:
        session = self._session_factory()
        if session.in_transaction():
            yield SqlAlchemyTransaction(session)
        else:
            with session.begin():
                yield SqlAlchemyTransaction(session)


class SqlAlchemyReadOnlyTransactionManager(SqlAlchemyTransactionManager):
    def __init__(self, engine: Engine) -> None:
        super().__init__(
            engine.execution_options(isolation_level="AUTOCOMMIT")
        )

    @contextmanager
    def transaction(self) -> t.Iterator[SqlAlchemyTransaction]:
        with self._session_factory() as session:
            yield SqlAlchemyTransaction(session)



import typing as t

from ...context import BaseContext
from ...result import Result
from ...unit import BaseUnit
from ...work import BaseWorkManager
from ...work.transactional import BaseTransactionManager
from ...work.transactional.base import TRANSACTION
from ..base import BaseUnitProxy

CONTEXT = t.TypeVar("CONTEXT", bound=BaseContext[t.Any])
OUT = t.TypeVar("OUT")


class TransactionalUnitProxy(BaseUnitProxy[CONTEXT, OUT]):
    def __init__(
        self,
        *,
        transaction_manager: BaseTransactionManager[TRANSACTION],
        unit: BaseUnit[CONTEXT, OUT],
    ) -> None:
        self._transaction_manager = transaction_manager
        self._unit = unit

    def __call__(self, context: CONTEXT) -> Result[OUT]:
        with self._transaction_manager.transaction() as trx:
            result = self._unit(context)

            if result.is_error():
                trx.rollback()
            else:
                trx.commit()

            return result


class TransactionalWorkManager(BaseWorkManager):
    def __init__(
        self,
        *,
        transaction_manager: BaseTransactionManager[TRANSACTION],
    ) -> None:
        self._transaction_manager = transaction_manager

    def by(self, unit: BaseUnit[CONTEXT, OUT]) -> BaseUnitProxy[CONTEXT, OUT]:
        return TransactionalUnitProxy(
            transaction_manager=self._transaction_manager,
            unit=unit,
        )



from .base import (
    BaseDataPointConsumerContext,
    BaseDataPointContext,
    BaseDataPointProducerContext,
)

__all__ = (
    "BaseDataPointConsumerContext",
    "BaseDataPointContext",
    "BaseDataPointProducerContext",
)



import typing as t
from datetime import datetime
from uuid import UUID

try:
    from sqlalchemy import UUID as SA_UUID
    from sqlalchemy import DateTime, Integer
    from sqlalchemy.orm import (
        DeclarativeBase,
        Mapped,
        MappedAsDataclass,
        mapped_column,
    )
except ImportError:  # pragma: no cover
    raise ImportError(
        "Seems that you are trying to import extra module that was not installed,"
        " please install pyuow[sqlalchemy]"
    )

TABLE_ID = t.TypeVar("TABLE_ID", bound=t.Any)


class BaseTable(MappedAsDataclass, DeclarativeBase):
    __abstract__ = True
    type_annotation_map = {
        UUID: SA_UUID(as_uuid=True),
    }


class EntityTable(BaseTable):
    __abstract__ = True
    id: Mapped[UUID] = mapped_column(SA_UUID, primary_key=True)


class AuditedEntityTable(EntityTable):
    __abstract__ = True
    created_date: Mapped[datetime] = mapped_column(
        DateTime(timezone=False), nullable=False
    )
    updated_date: Mapped[datetime] = mapped_column(
        DateTime(timezone=False), nullable=False
    )


class VersionedEntityTable(EntityTable):
    __abstract__ = True
    version: Mapped[int] = mapped_column(Integer(), nullable=False)



import abc
import typing as t
from abc import ABC

from ..context import BaseContext
from ..result import Result
from ..unit import BaseUnit

CONTEXT = t.TypeVar("CONTEXT", bound=BaseContext[t.Any])
OUT = t.TypeVar("OUT")


class BaseUnitProxy(BaseUnit[CONTEXT, OUT], ABC):
    def do_with(self, context: CONTEXT) -> Result[OUT]:
        return self(context)


class BaseWorkManager(ABC):
    @abc.abstractmethod
    def by(self, unit: BaseUnit[CONTEXT, OUT]) -> BaseUnitProxy[CONTEXT, OUT]:
        raise NotImplementedError



from .impl import (
    SqlAlchemyReadOnlyTransactionManager,
    SqlAlchemyTransaction,
    SqlAlchemyTransactionManager,
)

__all__ = (
    "SqlAlchemyReadOnlyTransactionManager",
    "SqlAlchemyTransaction",
    "SqlAlchemyTransactionManager",
)
